from __future__ import annotations

from pathlib import Path
from typing import Callable

from .models import FileOperation, ShowConfig


def plan_operations(
    show: ShowConfig,
    scanner: Callable[[Path], tuple[Path, object]],
) -> list[FileOperation]:

    operations: list[FileOperation] = []

    for file, match in scanner(show.path):

        show_name, season, episode, ep_name_raw, ext = match.groups()

        season_int: int = int(season)
        episode_int: int = int(episode)

        if season_int not in show.seasons:
            continue

        config = show.seasons[season_int]

        new_season: int = season_int + config.season_offset
        new_episode: int = episode_int + config.episode_offset

        if new_episode < 1:
            continue

        episode_name: str | None = (
            ep_name_raw.strip(" - ") if ep_name_raw else None
        )

        if episode_name:
            new_name: str = (
                f"{show_name} - S{new_season:02d}E{new_episode:02d} - {episode_name}.{ext}"
            )
        else:
            new_name = f"{show_name} - S{new_season:02d}E{new_episode:02d}.{ext}"

        new_dir: Path = show.path / f"Season {new_season}"
        new_path: Path = new_dir / new_name

        operations.append(
            FileOperation(
                source=file,
                destination=new_path,
            )
        )

    return operations


def detect_collisions(operations: list[FileOperation]) -> list[Path]:

    seen: set[Path] = set()
    collisions: list[Path] = []

    for op in operations:

        if op.destination in seen or op.destination.exists():
            collisions.append(op.destination)

        seen.add(op.destination)

    return collisions


def detect_duplicates(operations: list[FileOperation]) -> list[tuple[Path, Path]]:

    seen: dict[Path, Path] = {}
    duplicates: list[tuple[Path, Path]] = []

    for op in operations:

        if op.destination in seen:
            duplicates.append((seen[op.destination], op.source))
        else:
            seen[op.destination] = op.source

    return duplicates



from __future__ import annotations

from pathlib import Path
import typer
from rich import print

from .config import load_config, find_default_config
from .scanner import scan_show
from .planner import (
    plan_operations,
    detect_collisions,
    detect_duplicates,
)
from .preview import show_preview
from .executor import (
    execute_operations,
    get_default_undo_file,
    find_latest_undo_file,
)
from .undo import run_undo
from .doctor import run_doctor
from .cleanup import cleanup_shows

app = typer.Typer()


@app.command()
def run(
    config: Path | None = typer.Option(
        None,
        "--config",
        "-c",
        help="Config file (defaults to XDG config directory)",
    ),
    undo_file: Path | None = typer.Option(
        None,
        "--undo-file",
        help="Undo log file (defaults to XDG state directory)",
    ),
) -> None:
    """
    Apply episode shifts defined in config.
    """

    if config is None:
        config = find_default_config()

    cfg = load_config(config)

    all_operations = []

    for show in cfg.shows:

        operations = plan_operations(show, scan_show)

        show_preview(show, operations)

        all_operations.extend(operations)

    collisions = detect_collisions(all_operations)

    if collisions:
        print("[red]Collision detected:[/]")
        for c in collisions:
            print(c)
        raise typer.Exit(1)

    duplicates = detect_duplicates(all_operations)

    if duplicates:
        print("[yellow]Duplicate destination detected:[/]")
        for a, b in duplicates:
            print(a, b)

    if not typer.confirm("Apply these changes?"):
        raise typer.Exit()

    if undo_file is None:
        undo_file = get_default_undo_file()

    execute_operations(all_operations, undo_file)

    removed = cleanup_shows(show.path for show in cfg.shows)

    if removed:
        print(f"[blue]Removed empty folders:[/] {len(removed)}")

    print(f"[green]Done.[/] Undo file saved to: {undo_file}")


@app.command()
def undo(
    undo_file: Path | None = typer.Argument(
        None,
        help="Undo file (defaults to latest operation)",
    ),
) -> None:
    """
    Undo the latest rename operation.
    """

    if undo_file is None:
        undo_file = find_latest_undo_file()

        if undo_file is None:
            print("[red]No undo files found.[/]")
            raise typer.Exit(1)

        print(f"[blue]Using latest undo file:[/] {undo_file}")

    run_undo(undo_file)

    print("[green]Undo complete[/]")


@app.command()
def doctor(
    config: Path | None = typer.Option(
        None,
        "--config",
        "-c",
        help="Config file (defaults to XDG config directory)",
    ),
) -> None:
    """
    Run diagnostics on the library and configuration.
    """

    if config is None:
        config = find_default_config()

    cfg = load_config(config)

    run_doctor(cfg.shows)



from __future__ import annotations

from pathlib import Path
from pydantic import BaseModel


class SeasonConfig(BaseModel):
    season_offset: int = 0
    episode_offset: int = 0


class ShowConfig(BaseModel):
    name: str
    path: Path
    seasons: dict[int, SeasonConfig]


class Config(BaseModel):
    shows: list[ShowConfig]


class FileOperation(BaseModel):
    source: Path
    destination: Path


class UndoEntry(BaseModel):
    source: Path
    destination: Path



from __future__ import annotations

import json
from pathlib import Path
from typing import List

from .models import UndoEntry


def run_undo(undo_file: Path) -> None:

    data = json.loads(undo_file.read_text())

    entries: List[UndoEntry] = [
        UndoEntry.model_validate(e) for e in data
    ]

    # reverse order for safety
    for entry in reversed(entries):

        if not entry.source.exists():
            continue

        # recreate directory if cleanup removed it
        entry.destination.parent.mkdir(parents=True, exist_ok=True)

        entry.source.rename(entry.destination)



from __future__ import annotations

from pathlib import Path
from typing import Iterable


def remove_empty_dirs(root: Path) -> list[Path]:
    """
    Remove empty directories recursively under root.

    Returns list of removed directories.
    """

    removed: list[Path] = []

    # walk deepest directories first
    for path in sorted(root.rglob("*"), key=lambda p: len(p.parts), reverse=True):

        if path.is_dir():

            try:
                next(path.iterdir())
            except StopIteration:
                path.rmdir()
                removed.append(path)

    return removed


def cleanup_shows(show_paths: Iterable[Path]) -> list[Path]:

    removed: list[Path] = []

    for path in show_paths:
        removed.extend(remove_empty_dirs(path))

    return removed



from __future__ import annotations

import os
from datetime import datetime
from pathlib import Path
from typing import Optional

from .models import FileOperation, UndoEntry


APP_NAME: str = "seasonal-shift"


def get_state_dir() -> Path:
    xdg_state_home: str | None = os.getenv("XDG_STATE_HOME")

    if xdg_state_home:
        return Path(xdg_state_home) / APP_NAME

    return Path.home() / ".local" / "state" / APP_NAME


def get_default_undo_file() -> Path:
    state_dir: Path = get_state_dir()

    state_dir.mkdir(parents=True, exist_ok=True)

    timestamp: str = datetime.now().strftime("%Y%m%d-%H%M%S")

    return state_dir / f"undo-{timestamp}.json"


def find_latest_undo_file() -> Optional[Path]:

    state_dir: Path = get_state_dir()

    if not state_dir.exists():
        return None

    undo_files = sorted(
        state_dir.glob("undo-*.json"),
        key=lambda p: p.stat().st_mtime,
        reverse=True,
    )

    if not undo_files:
        return None

    return undo_files[0]


def execute_operations(
    operations: list[FileOperation],
    undo_file: Path,
) -> None:

    undo_entries: list[UndoEntry] = []

    for op in operations:

        op.destination.parent.mkdir(parents=True, exist_ok=True)

        op.source.rename(op.destination)

        undo_entries.append(
            UndoEntry(
                source=op.destination,
                destination=op.source,
            )
        )

    undo_file.write_text(
        "[\n"
        + ",\n".join(e.model_dump_json(indent=2) for e in undo_entries)
        + "\n]"
    )
